#define _GNU_SOURCE
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

#define START_CAPACITY  20
#define START_BUFSIZE  0x18
#define LINEBUF_SIZE   0x4000

enum calc_valtype {
	STRING,
	IDENTIFIER,
	NUM
};

struct calc_state_t;

typedef struct {
	union {
		char* str;
		int64_t num;
	};
	enum calc_valtype type;
	struct calc_state_t* state;
} CalcValue;

typedef struct {
	CalcValue* vals;
	size_t stkpointer;
	size_t capacity;
} CalcStack;

typedef void (*calc_fun)(struct calc_state_t*);

typedef struct {
	const char* name;
	size_t numArgs;
	calc_fun fun;
} FuncDefinition;

typedef struct {
	char* name;
	CalcValue val;
} VariableDefinition;

typedef struct calc_state_t {
	CalcStack stk;

	VariableDefinition* vars;
	size_t numVars;

	FuncDefinition* funcs;
	size_t numFuncs;

	bool isError;
} CalcState;

void init(void)
{
	int fd;
	unsigned int rand_val;
	// setting stdout and stdin to be unbuffered
	setvbuf(stdout,0,_IONBF,0);
	setvbuf(stdin,0,_IONBF,0);
	// seeding the random number generator
	fd = open("/dev/urandom", O_RDONLY);
	read(fd, &rand_val, sizeof(rand_val));
	close(fd);
	srand(rand_val);


	/* Retrieving this flag is the goal! */
	static char flag[0x40] = {0};
	FILE* flagfile = popen("/bin/get_flag", "r");
	fread(flag, sizeof(flag) - 1, 1, flagfile);
	pclose(flagfile);
}


/****** Variable storage******/

VariableDefinition* get_or_create_var(CalcState* state, const char* name)
{
	for(size_t i = 0; i < state->numVars; ++i)
	{
		if(!strcmp(state->vars[i].name, name))
			return &state->vars[i];
	}

	/* Not found, add one var */
	state->vars = realloc(state->vars, ++state->numVars * sizeof(*state->vars));

	VariableDefinition* r = &state->vars[state->numVars-1];
	r->name = strdup(name);
	return r;
}

VariableDefinition* get_var(const CalcState* state, const char* name)
{
	for(size_t i = 0; i < state->numVars; ++i)
	{
		if(!strcmp(state->vars[i].name, name))
			return &state->vars[i];
	}
	return NULL;
}

FuncDefinition* get_func(const CalcState* state, const char* name)
{
	for(size_t i = 0; i < state->numFuncs; ++i)
	{
		if(!strcmp(state->funcs[i].name, name))
			return &state->funcs[i];
	}
	return NULL;
}

/****** Variable storage ******/


/****** Stack ops ******/

void init_stack(CalcStack* stk)
{
	stk->vals = malloc(sizeof(CalcValue) * START_CAPACITY);
	stk->capacity = START_CAPACITY;
	stk->stkpointer = 0;
}

void push_stack(CalcStack* stk, const CalcValue* val)
{
	if(stk->stkpointer == stk->capacity)
	{
		stk->capacity *= 2;
		stk->vals = realloc(stk->vals, sizeof(CalcValue) * stk->capacity);
	}
	stk->vals[stk->stkpointer++] = *val;
}

void pop_stack(CalcStack* stk, CalcValue* rVal)
{
	*rVal = stk->vals[--stk->stkpointer];
}

bool pop_state_stack(CalcState* state, CalcStack* stk, CalcValue* rVal)
{
	CalcValue r;
	pop_stack(stk, &r);

	if(r.type == IDENTIFIER)
	{
		VariableDefinition* v = get_var(state, r.str);
		if(!v)
		{
			printf("Error: Undefined variable %s\n", r.str);
			state->isError = true;
			return false;
		}
		r = v->val;
	}
	*rVal = r;
	return true;
}

bool peek_state_stack(CalcState* state, CalcStack* stk, CalcValue* rVal)
{
	CalcValue r = stk->vals[stk->stkpointer - 1];

	if(r.type == IDENTIFIER)
	{
		VariableDefinition* v = get_var(state, r.str);
		if(!v)
		{
			printf("Error: Undefined variable %s\n", r.str);
			state->isError = true;
			return false;
		}
		r = v->val;
	}
	*rVal = r;
	return true;
}

void clean_stack(CalcStack* stk)
{
	for(size_t i = 0; i < stk->stkpointer; ++i)
	{
		if(stk->vals[i].type == STRING || stk->vals[i].type == IDENTIFIER)
			free(stk->vals[i].str);
	}
	stk->stkpointer = 0;
}

/****** END Stack ops ******/


/****** FUNCTIONS ******/

#define FUNC_NAME(name) func_##name
#define FUNC_NUM_OP(op, name) \
	void FUNC_NAME(name)(CalcState* state) \
	{ \
		CalcStack* stk = &state->stk; \
		CalcValue val1, val2; \
		pop_state_stack(state, stk, &val2); \
		pop_state_stack(state, stk, &val1); \
     \
		val1.num = val1.num op val2.num; \
		push_stack(stk, &val1); \
	}

FUNC_NUM_OP(+, add)
FUNC_NUM_OP(-, subtract)
FUNC_NUM_OP(*, multiply)
FUNC_NUM_OP(/, divide)
FUNC_NUM_OP(&, and)
FUNC_NUM_OP(|, or)
FUNC_NUM_OP(^, xor)

void func_pow(CalcState* state)
{
	CalcStack* stk = &state->stk;
	CalcValue val1, val2;
	pop_state_stack(state, stk, &val2);
	pop_state_stack(state, stk, &val1);

	int64_t tmp = 1;
	uint64_t exp = val2.num;
	for(;;)
	{
		if(exp & 1)
			tmp *= val1.num;
		exp >>= 1;

		if(!exp)
			break;
		val1.num *= val1.num;
	}
	val1.num = tmp;
	push_stack(stk, &val1);
}

void func_concatenate(CalcState* state)
{
	CalcStack* stk = &state->stk;
	CalcValue val1, val2, result;
	pop_state_stack(state, stk, &val2);
	pop_state_stack(state, stk, &val1);

	size_t len1 = strlen(val1.str);
	size_t addedLength = len1 + strlen(val2.str);
	char* newStr = malloc(addedLength+1);

	strcpy(newStr, val1.str);
	strcpy(newStr+len1, val2.str);

	result.type = STRING;
	result.str = newStr;
	result.state = val1.state;

	push_stack(stk, &result);

	free(val1.str);
	free(val2.str);
}

void func_rand(CalcState* state)
{
	CalcStack* stk = &state->stk;
	CalcValue val;
	val.state = state;
	val.type = NUM;
	val.num = 0;

	for(int i = 0; i < 4; ++i)
		val.num |= ((uint64_t) rand() & 0xffff) << 16 * i;

	push_stack(stk, &val);
}

void func_print(CalcState* state)
{
	CalcStack* stk = &state->stk;
	CalcValue val;
	peek_state_stack(state, stk, &val);

	switch(val.type)
	{
		case STRING:
			puts(val.str);
			break;

		case NUM:
			printf("%zd\n", val.num);
			break;
		default:
			printf("Error: Invalid type to be printed\n");
			state->isError = true;
			return;
	}
}

void func_toStr(CalcState* state)
{
	CalcStack* stk = &state->stk;
	CalcValue val;
	pop_state_stack(state, stk, &val);

	switch(val.type)
	{
		case STRING:
			break;

		case NUM:
		{
			char* newStr = malloc(0x18);
			snprintf(newStr, 18, "%zd", val.num);
			val.str = newStr;
			val.type = STRING;
			break;
		}

		default:
			printf("Error: Invalid type to be converted\n");
			state->isError = true;
			return;
	}
	push_stack(stk, &val);
}


void func_toNum(CalcState* state)
{
	CalcStack* stk = &state->stk;
	CalcValue val;
	pop_state_stack(state, stk, &val);

	switch(val.type)
	{
		case STRING:
		{
			int64_t num;
			char* end;
			num = strtoll(val.str, &end, 0);
			if(*end != 0)
			{
				printf("Error: Invalid string to be converted\n");
				state->isError = true;
				free(val.str);
				return;
			}
			free(val.str);
			val.type = NUM;
			val.num = num;
			break;
		}

		case NUM:
			break;

		default:
			printf("Error: Invalid type to be converted\n");
			state->isError = true;
			return;
	}
	push_stack(stk, &val);
}

void func_assign(CalcState* state)
{
	CalcStack* stk = &state->stk;
	CalcValue var, val;
	pop_state_stack(state, stk, &val);
	pop_stack(stk, &var);

	if(var.type != IDENTIFIER)
	{
		printf("Error: Trying to store to something that is not an identifier!\n");
		state->isError = true;
		push_stack(stk, &val); /* So it can be properly cleaned up later on */
		return;
	}

	get_or_create_var(state, var.str)->val = val;
}


void func_dup(CalcState* state)
{
	CalcStack* stk = &state->stk;
	CalcValue val, tmp;
	pop_stack(stk, &val);

	switch(val.type)
	{
		case NUM:
			push_stack(stk, &val);
			break;

		case STRING:
		case IDENTIFIER:
			tmp.type = val.type;
			tmp.str = strdup(val.str);
			tmp.state = val.state;
			push_stack(stk, &tmp);
			break;
	}

	push_stack(stk, &val);
}


FuncDefinition calc_funcs[] = {
	{"+", 2, func_add},
	{"-", 2, func_subtract},
	{"*", 2, func_multiply},
	{"/", 2, func_divide},
	{"&", 2, func_and},
	{"|", 2, func_or},
	{"^", 2, func_xor},
	{"**", 2, func_pow},
	{".", 2, func_concatenate},
	{"rand", 0, func_rand},
	{"print", 1, func_print},
	{"toStr", 1, func_toStr},
	{"toNum", 1, func_toNum},
	{"=", 2, func_assign},
	{"dup", 1, func_dup},
};

#define NUM_CALC_FUNCS (sizeof(calc_funcs) / sizeof(calc_funcs[0]))

/****** END FUNCTIONS ******/


/****** Eval logic ******/


bool parse_string(const char* tok, size_t* rlen, CalcValue* rVal)
{
#define ADD_CHAR(c) \
	if(curLen == capacity) \
	{ \
		capacity = capacity * 2 + 0x8; \
		buf = realloc(buf, capacity+1); \
	} \
	buf[curLen++] = *i; \

	bool escaped = false;

	char* buf = malloc(START_BUFSIZE + 1);
   size_t capacity = START_BUFSIZE;
   size_t curLen = 0;

	const char* begin = tok, *i;
	for(i = begin+1; *i; ++i)
	{
		if(escaped)
		{
			switch(*i)
			{
				case 'n':
					ADD_CHAR('\n');
					break;

				case 't':
					ADD_CHAR('\t');
					break;

				case 'r':
					ADD_CHAR('\r');
					break;

				case 'b':
					ADD_CHAR('\b');
					break;

				default:
					ADD_CHAR(*i);
					break;
			}
		}
		else if(*i == '"')
		{
			break;
		}
		else
		{
			ADD_CHAR(*i);
		}
	}
#undef ADD_CHAR

	if(*i != '"')
	{
		printf("Error: Unterminated string literal!\n");
		free(buf);
		return false;
	}
	buf[curLen] = 0;

	*rlen = i - tok + 1;
	rVal->type = STRING;
	rVal->str = buf;
	return true;

}

bool parse_number(const char* tok, size_t* rlen, CalcValue* rVal)
{
	char* end;

	int64_t num = strtoll(tok, &end, 0);
	*rlen = end - tok;

	rVal->type = NUM;
	rVal->num = num;

	return true;
}

bool parse_id(const char* tok, size_t* rlen, CalcValue* rVal)
{
	const char* i;
	for(i = tok ;*i && !isspace(*i); ++i);

	*rlen = i - tok;
	rVal->type = IDENTIFIER;
	rVal->str = strndup(tok, *rlen);
	return true;
}

bool parse_value(const char* tok, size_t* rlen, CalcValue* rVal)
{
	if(*tok == '"')
		return parse_string(tok, rlen, rVal);
	else if(isdigit(*tok))
		return parse_number(tok, rlen, rVal);
	else
		return parse_id(tok, rlen, rVal);
}

void eval_line(CalcState* state, const char* line)
{
	FuncDefinition* f;
	size_t rlen;
	CalcValue rVal;
	rVal.state = state;
	while(*line)
	{
		if(state->isError)
		{
			clean_stack(&state->stk);
			state->isError = false;
			return;
		}

		for(;*line && isspace(*line); ++line);

		if(!*line)
			break;

		if(parse_value(line, &rlen, &rVal))
		{
			if(rVal.type == IDENTIFIER && (f = get_func(state, rVal.str)))
			{
				free(rVal.str);
				if(state->stk.stkpointer < f->numArgs)
				{
					printf("Error: not enough arguments for function %s\n", f->name);
					state->isError = true;
					continue;
				}
				f->fun(state);
			}
			else
			{
				push_stack(&state->stk, &rVal);
			}
			line += rlen;
		}
		else
			state->isError = true;
	}

	while(state->stk.stkpointer)
	{
		func_print(state);
		pop_stack(&state->stk, &rVal);
		if(rVal.type == STRING || rVal.type == IDENTIFIER)
			free(rVal.str);
	}
}

/****** END Eval logic ******/

void print_welcome()
{
	printf("Welcome to our new cool stack-based custom calculator!\n"
			 "We use post-fix syntax, i.e. to compute 3+3, you would type \"3 3 +\"!\n"
			 "We also support variables, you can define them with the = operator, e.g. (\"a 3 =\").\n"
			 "To top it off: we also support strings, specifically:\n"
			 "\t- . Operator for concatenation (taken from awesome PHP!)\n"
			 "\t- toStr operator for converting other values to a string\n"
			 "\t- toNum for converting strings to numbers again\n"
			 );
}


void init_calcstate(CalcState* state)
{
	init_stack(&state->stk);

	state->vars = NULL;
	state->numVars = 0;

	state->funcs = calc_funcs;
	state->numFuncs = NUM_CALC_FUNCS;

	state->isError = false;
}

int main()
{
	char lineBuf[LINEBUF_SIZE];
	CalcState state;

	init();
	init_calcstate(&state);
	print_welcome();

	while (true)
	{
		printf("> ");
		if(!fgets(lineBuf, sizeof(lineBuf), stdin))
			return 0;

		eval_line(&state, lineBuf);
	}

}
